use crate::{
	physics::Cartesian, physics::Polar, physics::WorldPosition, sprite::SpriteAtlas, Doodad,
	DrawContext, Drawable,
};
use rand::Rng;
use sdl2::{gfx::primitives::DrawRenderer, pixels::Color};

struct StarField {
	stars: Vec<WorldPosition>,
	width: f64,
	height: f64,
}

impl StarField {
	fn new(width: f64, height: f64, d: f64) -> Self {
		let mut s = StarField {
			stars: Vec::new(),
			width,
			height,
		};

		let mut num = rand::thread_rng();
		let area = (width / 100.0) * (height / 100.0);
		let n = (area * d) as i32;
		for _ in 0..n {
			s.stars.push(WorldPosition::new(
				Cartesian::new(
					num.gen_range(-width / 2.0..width / 2.0),
					num.gen_range(-height / 2.0..height / 2.0),
				),
				num.gen_range(3f64..4f64),
			))
		}

		return s;
	}

	fn draw_with_offset(
		&self,
		dc: &DrawContext,
		pos_in_field: Cartesian,
		offset: Cartesian,
	) -> Result<(), String> {
		for wp in &self.stars {
			// Coordinate of star on screen,
			// with (0, 0) at top left
			let p: Cartesian = wp.pos.into();
			let q =
				((p - pos_in_field + offset) / wp.par - dc.top_left) * Cartesian::new(1.0, -1.0);

			dc.canvas.filled_circle(
				q.x as i16,
				q.y as i16,
				(5.0 - (1.0 * wp.par)) as i16,
				Color::RGBA(255, 255, 255, 100),
			)?;
		}
		return Ok(());
	}

	fn draw(&mut self, dc: &mut DrawContext, _sa: &SpriteAtlas) -> Result<(), String> {
		let w = self.width;
		let h = self.height;
		let ww = w / 2.0;
		let hh = h / 2.0;

		let pos_in_field = Cartesian::new(
			dc.camera.pos.x.signum() * (((dc.camera.pos.x.abs() + ww) % w) - ww),
			dc.camera.pos.y.signum() * (((dc.camera.pos.y.abs() + hh) % h) - hh),
		);

		// Field center in world coordinates
		let field_center = dc.camera.pos - pos_in_field;

		// screen coordinates of top left position of starfield
		let field_tl =
			WorldPosition::new(field_center + Cartesian::new(-ww, hh), 3.0).screen_position(dc);

		let field_br =
			WorldPosition::new(field_center + Cartesian::new(ww, -hh), 3.0).screen_position(dc);

		dc.canvas.aa_circle(
			field_tl.x as i16,
			field_tl.y as i16,
			5,
			Color::RGB(255, 0, 0),
		)?;
		dc.canvas.set_draw_color(Color::RGB(0, 0, 0));

		let north = field_tl.y > 0.0;
		let south = field_br.y < dc.window_size.y;
		let east = field_br.x < dc.window_size.x;
		let west = field_tl.x > 0.0;

		self.draw_with_offset(dc, pos_in_field, Cartesian::new(0.0, 0.0))?;
		if north {
			self.draw_with_offset(dc, pos_in_field, Cartesian::new(0.0, h))?;
		}
		if south {
			self.draw_with_offset(dc, pos_in_field, Cartesian::new(0.0, -h))?;
		}
		if east {
			self.draw_with_offset(dc, pos_in_field, Cartesian::new(w, 0.0))?;
		}
		if west {
			self.draw_with_offset(dc, pos_in_field, Cartesian::new(-w, 0.0))?;
		}
		if north && east {
			self.draw_with_offset(dc, pos_in_field, Cartesian::new(w, h))?;
		}
		if north && west {
			self.draw_with_offset(dc, pos_in_field, Cartesian::new(-w, h))?;
		}
		if south && east {
			self.draw_with_offset(dc, pos_in_field, Cartesian::new(w, -h))?;
		}
		if south && west {
			self.draw_with_offset(dc, pos_in_field, Cartesian::new(-w, -h))?;
		}

		dc.canvas.set_draw_color(Color::RGB(0, 0, 0));
		return Ok(());
	}
}

pub struct System {
	bodies: Vec<Box<dyn Drawable>>,
	starfield: StarField,
}

impl System {
	pub fn new() -> Self {
		let mut s = System {
			bodies: Vec::new(),
			starfield: StarField::new(6000.0, 6000.0, 1.0),
		};

		s.bodies.push(Box::new(Doodad {
			pos: WorldPosition::new(Cartesian::new(0.0, 0.0), 2.0),
			sprite: "a0.png".to_owned(),
			scale: 1,
			angle: 0.0,
		}));

		s.bodies.push(Box::new(Doodad {
			pos: WorldPosition::new(
				Polar {
					center: Cartesian::new(0.0, 0.0),
					radius: 300.0,
					angle: 31.0,
				}
				.into(),
				1.5,
			),
			sprite: "earth.png".to_owned(),
			scale: 1,
			angle: (180f64).to_radians(),
		}));

		s.bodies.push(Box::new(Doodad {
			pos: WorldPosition::new(Cartesian::new(1000.0, 1000.0), 2.0),
			sprite: "small.png".to_owned(),
			scale: 1,
			angle: 0.0,
		}));

		return s;
	}

	/// Calculate the state of this body after t seconds.
	pub fn tick(&mut self, _t: f64) {
		//let body = &mut self.bodies[1];
		//body.pos.angle += 0.1 * t;
		//body.angle -= 0.1 * t;
	}

	pub fn draw(&mut self, dc: &mut DrawContext, sa: &SpriteAtlas) -> Result<(), String> {
		self.starfield.draw(dc, sa)?;
		for body in &self.bodies {
			body.draw(dc, sa)?;
		}

		return Ok(());
	}
}
